"use strict";
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var react_1 = __importStar(require("react"));
var lodash_1 = require("lodash");
var immer_1 = __importDefault(require("immer"));
var MISSING = {};
var defaultStateTreeAPI = {
    stateTree: null,
    replaceStateTree: function () { },
    markGeneratedPath: function () { },
    checkIfGeneratedPathInUse: function () { return false; },
};
var StateTree = react_1.default.createContext(defaultStateTreeAPI);
var StateTreePath = react_1.default.createContext([]);
function StateTreeProvider(_a) {
    var children = _a.children, 
    // @ts-ignore
    _b = _a.initialValue, 
    // @ts-ignore
    initialValue = _b === void 0 ? {} : _b, _c = _a.onUpdate, onUpdate = _c === void 0 ? function () { } : _c;
    var _d = __read(react_1.useState(initialValue), 2), stateTree = _d[0], rawReplaceStateTree = _d[1];
    var generatedPaths = react_1.useRef({});
    return (react_1.default.createElement(StateTree.Provider, { value: {
            stateTree: stateTree,
            replaceStateTree: function (nextStateTree) {
                rawReplaceStateTree(nextStateTree);
                onUpdate(nextStateTree);
            },
            markGeneratedPath: function (path) {
                lodash_1.set(generatedPaths.current, path, true);
            },
            checkIfGeneratedPathInUse: function (path) {
                return lodash_1.get(generatedPaths.current, path, false);
            },
        } }, children));
}
exports.StateTreeProvider = StateTreeProvider;
function StateKey(_a) {
    var value = _a.value, children = _a.children;
    var parentPath = react_1.useContext(StateTreePath);
    return (react_1.default.createElement(StateTreePath.Provider, { value: parentPath.concat(value) }, children));
}
exports.StateKey = StateKey;
function StateListKey(_a) {
    var value = _a.value, children = _a.children;
    return (react_1.default.createElement(StateKey, { value: value }, react_1.default.Children.map(children, function (child, index) {
        return (react_1.default.createElement(StateKey, { key: index, value: index }, child));
    })));
}
exports.StateListKey = StateListKey;
function useEntireStateTree() {
    var _a = react_1.useContext(StateTree), stateTree = _a.stateTree, replaceStateTree = _a.replaceStateTree;
    return { stateTree: stateTree, replaceStateTree: replaceStateTree };
}
exports.useEntireStateTree = useEntireStateTree;
var defaultStateTreeKey = null;
function nextStateTreeKey(nextKey) {
    defaultStateTreeKey = nextKey;
}
exports.nextStateTreeKey = nextStateTreeKey;
function useStateTree(initialValue, key) {
    if (key === void 0) { key = defaultStateTreeKey; }
    defaultStateTreeKey = null;
    var keyRef = react_1.useRef(key);
    var stateTreeApi = react_1.useContext(StateTree);
    if (stateTreeApi === defaultStateTreeAPI) {
        // If they aren't using a StateTreeProvider, fall back to useState
        return react_1.useState(initialValue);
    }
    var stateTree = stateTreeApi.stateTree, replaceStateTree = stateTreeApi.replaceStateTree, markGeneratedPath = stateTreeApi.markGeneratedPath, checkIfGeneratedPathInUse = stateTreeApi.checkIfGeneratedPathInUse;
    var stateTreePath = react_1.useContext(StateTreePath);
    if (keyRef.current == null) {
        var generatedKey = void 0;
        var index = 0;
        var valueAtStateTreePath = lodash_1.get(stateTree, stateTreePath, MISSING);
        if (valueAtStateTreePath !== MISSING) {
            index = Object.keys(valueAtStateTreePath).length;
        }
        do {
            generatedKey = "$" + index;
            index++;
        } while (checkIfGeneratedPathInUse(stateTreePath.concat(generatedKey)));
        markGeneratedPath(stateTreePath.concat(generatedKey));
        keyRef.current = generatedKey;
    }
    if (keyRef.current == null) {
        throw new Error("Expected keyRef to be initialized, but it was not");
    }
    var targetPath = stateTreePath.concat(keyRef.current);
    var updateState = function (nextValue) {
        replaceStateTree(immer_1.default(stateTree, function (draft) {
            lodash_1.set(draft, targetPath, nextValue);
        }));
    };
    var value = lodash_1.get(stateTree, targetPath, MISSING);
    if (value === MISSING) {
        updateState(initialValue);
        value = initialValue;
    }
    return [value, updateState];
}
exports.useStateTree = useStateTree;
